///////////////////////////////////////////////
//          scicosPreprocess.sci
// GOAL : apply some modifications to the data 
// given by the scicos precompiler in order to 
// have coherent and complete data
///////////////////////////////////////////////

// note: this part is non optimal for the sake of clarity. 
// We could only process clkMat once.

// note about the lines of the clk/data matrices :
// Mat(i,1) is the id of a block sending data/event by its regular/event output
// port Mat(i,2) to the block of id Mat(i,3) on its regular/event input port Mat(i,4)



//*** Main function, returns a list of blkInfo and the two matrices ***//
// clkMat and dataMat (possibly) modified
// It also does some coherency check on the structures
// args
// cpr, infos gathered by scicos for each block.
// clkMat dataMat, see above
// clkRootId, is the scicos id of the root bock
function [cpr, clkMat, dataMat, clkRootId] = scicosPreprocess (cpr, clkMat, dataMat)
// ******** INIT MAIN VARIABLES ************ //
clkRootId = 0;
// ******** END INIT MAIN VARIABLES ******** //

mprintf ("\n**gathering blocks infos\n");
mprintf ("\n**checking name for each block\n");

for i=1:cpr.sim.nb
  // filling name
  cpr.sim.funs(i) = C_getCoherentName(cpr, i);

  // storing clkRoot id for checkMatValidity
  if cpr.sim.funs(i) == "clkRoot" then 
    clkRootId = i;
    mprintf("\n**Event actvation root found (id:%d)\n", clkRootId);
  end
end

if clkRootId == 0 then error("no root found in this graph... this is very strange..."); end


mprintf("\n**force root activation for actuators :");
clkMat = forceActuatorRootActivation (cpr, clkMat, clkRootId);

mprintf("\n**clkconnect check about root activations :");
clkMat = rootActivationCheck (clkMat, clkRootId);

mprintf("\n**check if every block have at least 1 activation :");
clkMat = atLeastOneEventCheck (cpr, clkMat, clkRootId);

mprintf("\n**check for activation redundancy :");
clkMat = atMostOneEventCheck (clkMat);

// mprintf("\n**second check for multi activations :");
// ok = multiActivationCheck (clkMat);


mprintf("\n\n*************\nRECAP :\nNb blocks:%d\nnb activations:%d\nnb dependences:%d\n*************\n",...
    cpr.sim.nb, size(clkMat,'r'), size(dataMat,'r'));
endfunction

// force an actuator to be activated by root
// note we do not process multiactivation here since it is taken care of later
function clkMat = forceActuatorRootActivation (cpr, clkMat, clkRootId)

return; // MODIF 03/01/05

for i=1:size(clkMat,'r')
  clkLine = clkMat(i,:);
  
  // activated block is an actuator -> force root
  if C_blkIsActuator(cpr, clkLine(3)) then
    //    mprintf("Line:%d, blk %d : force to clkRoot\n",i,clkLine(3));
    clkMat(i, :) = [clkRootId, 1, clkLine(3), clkLine(4)]; 
  end
end
endfunction

// ensure that each block is activated once and only once
// if not activated, adding a clkLine to activate block by clkRoot
function clkMat = atLeastOneEventCheck(cpr, clkMat, clkRootId)

//mprintf("Checking if each block has at least 1 activation\n");
for i=1:cpr.sim.nb
  // looking for an activation for each block
  found = %F;
  j = 1; n = size(clkMat,'r');
  while j<=n & found==%F
    if i == clkMat(j,3) then
      found = %T;
    end
    j = j + 1;
  end
  
  
  // no activation AND block isn't clkRoot
  // => force 1 activation by root
  if ~found & cpr.sim.funs(i) <> "clkRoot" then
    mprintf(".");
//    str = sprintf ("%d(%s) | ",i, cpr.sim.funs(i));
//    mprintf(str);
    clkMat($+1, :) = [clkRootId, 1, i, 1];
  end
end
endfunction

// ensure that a block has at most one activation from a same (block,port)
// arg : clkMat, is the connection matrix (I/O)
//
// this function is probably incorrect when clkRoot has more than 1 event output port
// but since we only handle monorythm, everything should be ok
function clkMat = atMostOneEventCheck (clkMat)
// current clk matrix size
curSize = size(clkMat,'r');
i=1;

while i <= curSize
  j=i+1; //found=%F;
  
  // if there is many times the same line in clkMat, delete all but one
  while j <= curSize
    if clkMat(i,1) == clkMat(j,1) & clkMat(i,2) == clkMat(j,2) & clkMat(i,3) == clkMat(j,3) then 
      mprintf("[%d]",clkMat(i,3));
      
      clkMat(j,:) = []; // deleting line 'j'
      curSize = curSize -1;
    else
      j = j+1;
    end
  end
  
  // DEBUG
  // force the remaining block activating 'i' to be clkRoot
  //  if found then 
  //    str = sprintf("found more than 1 activation for block %d, reducing to clkRoot", i)
  //    mprintf(str);
  //    clkMat(i,:) = rootActivateLine (clkMat(i,:));
  //  end
  
  i = i+1;
end
endfunction


// for now, this is a debug function. It checks if a block is activated more than 1 time
// args:
// clkMat, blkInfos : as usual
// ok : a boolean
//
// -> this function fellows the same pattern than function atMostOneEventCheck
function ok = multiActivationCheck (clkMat)
curSize = size(clkMat,'r');
curMat = clkMat; // the clkMat we will process (and modify)
ok=%T;

blkLst=list(); // list of multiactivated blocks id
// it is a list of couple (blk,port) of activation 
// (index of the list is the block id -> blkLstPtr helps to recover info)
blkLstPtr=list(); // pointer list to blkLst


// process is cut into 2 phases in order to allow further code modifications
// an optimization would be to stop after the first multiactivated bad block

// 1st : detection, we build a list of id
i=1;
while i <= curSize
  currentBlkId = curMat(i,3);
  
  activationList = list(); // list of activation(s)
  activationList(1) = list(curMat(i,1),curMat(i,2)); // storing the first (clkBlk,clkPort)

  j=i+1;
  while j <= curSize
    // we assert that curMat(i,1)<>curMat(j,1) & curMat(i,2)<>curMat(j,2)
    if currentBlkId == curMat(j,3) then // multi activation found
      activationList($+1) = list(curMat(j,1),curMat(j,2)); // adding to the list
      curMat(j,:) = []; curSize = curSize-1; // deleting line in order to not re processing it
    else
      j = j+1;
    end
  end

  if size(activationList) > 1 then // if there is more than one activation
    blkLst(currentBlkId) = activationList; 
    blkLstPtr($+1)= currentBlkId; // pointer list to blkLst;
  end
  i = i+1;
end


// 2nd : we check for every block id if it is allowed
// at this point we have these variables: blkLst, blkLstPtr, blkInfos
// for now, blkLst is unused. We keep it for further processing

//for i=1:size(blkInfos)
//  mprintf("blk %d(%s) z=%d\n",i,blkInfos(i).name,blkInfos(i).zStates);
//end

//for blkId=blkLstPtr
//  nbDiscrSt = blkInfos(blkId).zStates;
//  blkName = blkInfos(blkId).name;
//  if nbDiscrSt > 0 then
//    str = sprintf("multi activated block %d(%s) is forbidden (has %d >0 discrete state)", ..
//	blkId, blkName, nbDiscrSt);
//    warning(str);
//    ok=%F;
//  end
//end

endfunction



// returns the sum of all elements in intList
// returns 0 if a negative/zero size is encoutered
function intSum = sumOfList (intList)
intSum = 0;
for i=intList
  if i <= 0 then
    intSum = 0;
    return;
  else
   intSum = intSum +i;
  end
end
endfunction




// Scicos create a dummy block with id '0' in order to simulate continuous time.
// Since we don't take care of it, we replace continuous activations by clkRoot activations
function clkMat = rootActivationCheck (clkMat, clkRootId)
for i=1:size(clkMat,'r')
  if clkMat(i,1) == 0 then
    mprintf(".");
    clkMat(i,1) = clkRootId;
    // warning, in complex models clkRoot may have more than one activation output...

    clkMat(i,2) = 1;
  end
end
endfunction
